@@ -50,7 +50,9 @@ from clusterondemand.node_definition import NodeDefinition
 from clusterondemand.ssh import clusterssh_ns
 from clusterondemand.ssh_key import validate_ssh_pub_key
 from clusterondemand.summary import SummaryType
+from clusterondemand.utils import parse_assignment
 from clusterondemand.wait_helpers import clusterwaiters_ns, wait_for_cluster_to_be_ready
+from clusterondemandazure.azure_actions.credentials import AzureApiHelper
 from clusterondemandazure.azure_actions.storage import StorageAction
 from clusterondemandazure.base import ClusterCommand
 from clusterondemandazure.inbound_traffic_rule import ARM_format
@@ -70,7 +72,7 @@ SUPPORTED_DISK_SKUS = ["Standard_LRS", "Premium_LRS", "StandardSSD_LRS", "UltraS
 config_ns = ConfigNamespace("azure.cluster.create", "cluster creation parameters")
 config_ns.import_namespace(clusterondemand.configuration.clustercreate_ns)
 config_ns.import_namespace(findimages_ns)
-config_ns.override_imported_parameter("version", default="9.1")
+config_ns.override_imported_parameter("version", default="9.2")
 config_ns.import_namespace(clusterondemand.configuration.cmd_debug_ns)
 config_ns.import_namespace(clusterssh_ns)
 config_ns.import_namespace(clusterwaiters_ns)
@@ -86,15 +88,21 @@ config_ns.add_parameter(
     choices=SUPPORTED_DISK_SKUS,
     default="StandardSSD_LRS",
     help="Storage type to use for Azure root volume")
-config_ns.override_imported_parameter("node_type", default="Standard_D1_v2")
-config_ns.override_imported_parameter("head_node_type", default="Standard_D1_V2")
+config_ns.override_imported_parameter(
+    "node_type",
+    default="Standard_D1_v2",
+    help="The instance type of compute nodes. It must exist in the region you use. "
+    "Can be suffixed with ':V1' or ':V2', meaning the Hyper-V generation. The default is V1.",
+)
+config_ns.override_imported_parameter(
+    "head_node_type",
+    default="Standard_D1_V2",
+    help="The instance type must exist in the region you use. "
+    "Can be suffixed with ':V1' or ':V2', meaning the Hyper-V generation. The default is V1.",
+)
 config_ns.override_imported_parameter(
     "ssh_pub_key_path", validation=lambda p, c: validate_ssh_pub_key(p, c, allowed_types={"RSA": 2048})
 )
-config_ns.add_parameter(
-    "vhd_url",
-    advanced=True,
-    help="The url for the source vhd for the server side copy")
 config_ns.add_parameter(
     "head_node_image",
     help=("Single image selector statement for the head node image. Can either be an image "
@@ -148,6 +156,19 @@ config_ns.add_parameter(
     help="Name of the subnet to use.",
     validation=requires_other_parameters_to_be_set(["vnet_resource_group", "vnet_network_name"]),
 )
+config_ns.add_enumeration_parameter(
+    "head_node_tags",
+    advanced=True,
+    parser=parse_assignment,
+    default=None,
+    help=("Tags which are to be configured to head node machine of the cluster "
+          "(i.e. --tags tag1=value tag2=value).")
+)
+config_ns.add_switch_parameter(
+    "accept_azure_terms",
+    advanced=True,
+    help="Accept the most recent Azure marketplace terms and skip the dialog"
+)
 
 
 def internal_subnet_id_validation(subnet_id_param, config):
@@ -210,6 +231,34 @@ HR = "---------------------------------------------------------------------"
 
 class ClusterCreate(ClusterCommand):
 
+    @staticmethod
+    def _get_instance_type(instance_type_and_vm_gen):
+        """ Extracts the instance type from a string like '<type>[:V<gen>]'. """
+        return instance_type_and_vm_gen.split(":")[0]
+
+    @staticmethod
+    def _get_vm_gen(instance_type_and_vm_gen):
+        """ Extracts the Hyper-V generation from a string like '<type>[:V<gen>]',
+            where <gen> is either 1 or 2. If the generation is missing, it defaults to 1. """
+        tokens = instance_type_and_vm_gen.split(":")
+        return tokens[1].upper() if len(tokens) > 1 else "V1"
+
+    @property
+    def head_node_type(self):
+        return self._get_instance_type(config["head_node_type"])
+
+    @property
+    def head_node_vm_gen(self):
+        return self._get_vm_gen(config["head_node_type"])
+
+    @property
+    def node_type(self):
+        return self._get_instance_type(config["node_type"])
+
+    @property
+    def node_vm_gen(self):
+        return self._get_vm_gen(config["node_type"])
+
     def _get_head_node_name(self, cluster_name):
         return "%s-head-node" % cluster_name
 
@@ -233,7 +282,7 @@ class ClusterCreate(ClusterCommand):
             "bright": {
                 "node_disk_setup_path": "/root/cm/disk-setup.xml",
                 "wlm": kwargs["wlm"],
-                "wlm_slot_count": 1,
+                "wlm_slot_count": "AUTO" if BcmVersion(config["version"]) > "9.1" else 1,
                 # None   type gets translated to 'None' string somewhere,
                 # that's why need need empty string here for pbsproc_lic_server
                 "pbspro_lic_server": "",
@@ -247,7 +296,7 @@ class ClusterCreate(ClusterCommand):
                 "region": self.location,
                 "resource_group": self.resource_group,
                 "storage_account": self.storage_account.name,
-                "node_type": config["node_type"],
+                "node_type": self.node_type,
                 "subscription_id": config["azure_subscription_id"],
                 "nodes": {
                     "count": config["nodes"],
@@ -256,12 +305,13 @@ class ClusterCreate(ClusterCommand):
                         "node-installer-disk": 2,
                     },
                     "base_name": "cnode",
-                    "type": config["node_type"],
+                    "type": self.node_type,
                 },
                 "client_id": config["azure_client_id"],
                 "client_secret": config["azure_client_secret"],
                 "tenant_id": config["azure_tenant_id"],
                 "network_cidr": str(config["subnet_cidr"]),
+                "default_hyperv_gen": self.node_vm_gen,
                 "head_node_ip": str(config["head_node_ip"])
             }
         }
@@ -337,7 +387,7 @@ class ClusterCreate(ClusterCommand):
 
         cloud_init_script = """#!/bin/bash
 
-        mkdir /root/.ssh/
+        mkdir -p /root/cm
 
         echo '%s' > /root/cm/cm-bright-setup.conf
         echo '%s' > /root/cm/disk-setup.xml
@@ -346,7 +396,7 @@ class ClusterCreate(ClusterCommand):
         if config["ssh_pub_key_path"]:
             with open(config["ssh_pub_key_path"]) as pub_key:
                 cloud_init_script += """
-                mkdir /root/.ssh/
+                mkdir -p /root/.ssh/
                 echo '%s' >> /root/.ssh/authorized_keys
                 """ % pub_key.read()
 
@@ -398,7 +448,8 @@ class ClusterCreate(ClusterCommand):
         self._validate_cluster_password()
         self._validate_access_credentials()
         self._validate_location()
-        self._validate_vmsize()
+        self._validate_vmsize(self.head_node_type, self.node_type)
+        self._validate_vm_gen(self.head_node_vm_gen, self.node_vm_gen)
         self._validate_blob()
 
     def on_error(self, r_client):
@@ -440,7 +491,7 @@ class ClusterCreate(ClusterCommand):
                 # The IP address here is arbitrary and was just added to satisfy Azure API
                 # requirements.
             except HttpResponseError as e:
-                log.debug(e.message)
+                AzureApiHelper.log_error_details(e)
                 raise UserReportableException(
                     "Insufficient write access: Not enough permissions to create a virtual network."
                 )
@@ -453,7 +504,7 @@ class ClusterCreate(ClusterCommand):
             async_removal.wait()
 
         except HttpResponseError as e:
-            log.debug(e.message)
+            AzureApiHelper.log_error_details(e)
             raise UserReportableException(
                 "Insufficient write access: unable to create a resource group."
             )
@@ -482,6 +533,7 @@ class ClusterCreate(ClusterCommand):
         except UserReportableException as e:
             raise e
         except HttpResponseError as e:
+            AzureApiHelper.log_error_details(e)
             self.on_error(self.azure_api.resource_client)
             raise CODException("Template deployment failed: %s" % str(e), caused_by=e)
         except Exception as e:
@@ -524,13 +576,16 @@ class ClusterCreate(ClusterCommand):
                     caching="ReadOnly",
                 )
             ),
-            hyper_v_generation=HyperVGenerationTypes.V1
+            hyper_v_generation=(HyperVGenerationTypes.V2
+                                if self.head_node_vm_gen == "V2"
+                                else HyperVGenerationTypes.V1)
         )
 
         creation_successful = False
         current_attempt = 0
         MAX_ATTEMPTS = 3
-        log.info("Creating Azure Image resource '%s' from blob %s", img_name, blob_url)
+        log.info("Creating Azure Image resource '%s' (Hyper-V gen %s) from blob %s",
+                 img_name, img_params.hyper_v_generation.value, blob_url)
         while current_attempt < MAX_ATTEMPTS and not creation_successful:
             try:
                 create_img_future = self.azure_api.compute_client.images.begin_create_or_update(
@@ -545,6 +600,7 @@ class ClusterCreate(ClusterCommand):
                     log.debug("Failed to create the headnode image, retrying...")
                     current_attempt += 1
                     continue
+                AzureApiHelper.log_error_details(e)
                 raise CODException("Failed to create the headnode image", caused_by=e)
 
     def check_resource_group_usability(self):
@@ -580,6 +636,7 @@ class ClusterCreate(ClusterCommand):
             try:
                 self.azure_api.compute_client.images.get(self.resource_group, head_node_image_name)
             except HttpResponseError as e:
+                AzureApiHelper.log_error_details(e)
                 raise CODException(f"Resource group '{self.resource_group}' doesn't have"
                                    f" the head node image '{head_node_image_name}'", caused_by=e)
         else:
@@ -608,13 +665,13 @@ class ClusterCreate(ClusterCommand):
         self.location = config["azure_location"]
         inbound_rules = self.generate_inbound_traffic_rules(config["inbound_rule"])
 
-        if config["vhd_url"]:
-            if config["head_node_image"]:
-                raise CODException("--vhd-url and --head-node-image cannot be used at the same time")
-            # For compatiblity with the existing vhd_url parameter
-            config["head_node_image"] = config["vhd_url"]
-
         self.head_node_image = AzureImageSource.pick_head_node_image_using_options(config)
+        if not self.head_node_image.version and config["run_cm_bright_setup"]:
+            log.warning(
+                f"Using custom image: {self.head_node_image.uuid} with parameter run_cm_bright_setup set to 'yes'."
+                f" Probably it was set by mistake because a custom image might not have necessary files"
+                f" to run cm-bright-setup. Consider using --run-cm-bright-setup=no to "
+            )
 
         self.use_existing_vnet = False
         if config["vnet_resource_group"]:
@@ -632,9 +689,11 @@ class ClusterCreate(ClusterCommand):
 
         generator = AzureSummaryGenerator(config,
                                           SummaryType.Proposal,
-                                          head_node_definition=NodeDefinition(1, config["head_node_type"]),
+                                          head_node_definition=NodeDefinition(
+                                              1, self.head_node_type + ":" + self.head_node_vm_gen),
                                           head_image=self.head_node_image,
-                                          node_definition=NodeDefinition(config["nodes"], config["node_type"]),
+                                          node_definition=NodeDefinition(
+                                              config["nodes"], self.node_type + ":" + self.node_vm_gen),
                                           region=self.location)
         generator.print_summary(log.info)
 
@@ -695,7 +754,7 @@ class ClusterCreate(ClusterCommand):
         utils.cod_log(log, "Building deployment template", 45)
         tb = TemplateBuilder(
             head_node_name=self.head_node,
-            head_node_flavour=config["head_node_type"],
+            head_node_flavour=self.head_node_type,
             storage_account=self.storage_account.name,
             region=self.location,
             custom_data=_b64encode(self.cloud_init_script()),
@@ -740,24 +799,12 @@ class ClusterCreate(ClusterCommand):
                 raise UserReportableException(
                     "The requested vmsize: '{vmsize}' is not available in the location: "
                     "'{location}' temporarily, please select a different region or vmsize.".format(
-                        vmsize=config["head_node_type"],
+                        vmsize=self.head_node_type,
                         location=config["azure_location"]
                     )
                 )
-            if "DeploymentFailed" in e.message:
-                try:
-                    for operation in self.azure_api.resource_client.deployment_operations.list(
-                        resource_group_name=self.resource_group,
-                        deployment_name=self.deployment_name
-                    ):
-                        if operation.properties.status_code.upper() != "OK":
-                            resource_name = operation.properties.target_resource.resource_name
-                            error_code = operation.properties.status_message.error.code
-                            error_message = operation.properties.status_message.error.message
-                            log.error(f"Deployment error for resource {resource_name}: ({error_code}) {error_message}")
-                except Exception as inner_e:
-                    log.error(f"Error while diagnosing deployment failure: {inner_e}")
-
+            AzureApiHelper.log_error_details(e, azure_api=self.azure_api, resource_group=self.resource_group,
+                                             deployment_name=self.deployment_name)
             self.on_error(self.azure_api.resource_client)
             raise CODException("Template deployment failed: %s" % e.message, caused_by=e)
         except Exception as e:
@@ -774,7 +821,10 @@ class ClusterCreate(ClusterCommand):
             ).ip_address
             log.info("Head node IP: %s" % public_ip)
 
-            wait_for_cluster_to_be_ready(config, public_ip, config["version"])
+            if config["run_cm_bright_setup"]:
+                # Normally disabling run_cm_bright_setup automatically disables waiting for cluster,
+                # but in case of custom vhd url this parameter is overridden after validation.
+                wait_for_cluster_to_be_ready(config, public_ip, config["version"])
         else:
             public_ip = "N/A"
             log.info("Cluster was created without a public IP. COD cannot wait for cmdaemon to be ready.")
