@@ -23,6 +23,7 @@ from azure.core.exceptions import HttpResponseError
 import clusterondemand.clustercreate
 from clusterondemand.codoutput.sortingutils import SortableData
 from clusterondemand.exceptions import UserReportableException
+from clusterondemandazure.azure_actions.credentials import AzureApiHelper
 from clusterondemandazure.base import ClusterCommand
 from clusterondemandconfig import ConfigNamespace, config
 
@@ -73,6 +74,7 @@ class ClusterList(ClusterCommand):
         except HttpResponseError as e:
             if "SubscriptionNotFound" in e.error.code:
                 raise UserReportableException(str(e))
+            AzureApiHelper.log_error_details(e)
             raise e
 
         rows = []
